/*
   *Copyright:  

   ACEMod (2018--2021), The University of Sydney's invention disclosure CDIP Ref. 2019-123.
   Copyright ©  Oliver Cliff, Cameron Zachreson, Nathan Harding, Kristopher Fair, Mikhail Prokopenko, Mahendarajah Piraveenan

   AMTraC-19 (2020-2021), The University of Sydney's invention disclosure CDIP 2020-018 (2020).
   Copyright ©  Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   AMTraC-19 (2022-2023), The University of Sydney.
   Copyright ©  Sheryl Chang, Quang Dang Nguyen, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   AMTraC-19 (2024), The University of Sydney.
   Copyright ©  Quang Dang Nguyen, Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   *EndCopyright:
*/

// Compile the code: g++ generate_population_2021.cpp -std=c++11 -O2
// Check our userguide for more information on how to run the code.

#include <limits.h>
#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <vector>
#include <stdarg.h>
#include <math.h>
#include <string>
#include <algorithm>
#include <map>
#include <sstream>

#include <chrono>
#include <thread>


#include "debug.h"
#include "utils.hpp"
#include "generate_population_2021.h"
#include "enums.h"

void generatePopulation( std::vector< std::shared_ptr< SimpleAgent > >& );

void printPopulation( const std::vector< std::shared_ptr< SimpleAgent > >&, const std::string& );

void school_configuration( schoolsize_school_map_type&, school_postcode_map_type&,
	school_schoolsize_map_type&, postcode_sla_map_type&, sla_numStudents_map_type&,
	sla_dz_map_type&, dz_numAgents_map_type&, numStudents_sla_map_type&,
	sla_school_map_type&, school_id_map_type&);

	std::shared_ptr< SimpleAgent > generateAgent( unsigned long, cd_type, unsigned long, unsigned long, const std::vector< std::pair< AgeGroup, float > >& );

	int main(int argc, char *argv[])
	{
		srand(time(NULL));

		std::vector< std::shared_ptr< SimpleAgent > > agents;
		const std::string fname = output_directory + "population.dat";
		std::cout << "Initiating program." << std::endl;

		std::ofstream ACARA_Discrepency(output_directory + "ACARA_Discrepency.dat");
		ACARA_Discrepency << "postcode" << " " << "school_size" << std::endl;
		ACARA_Discrepency.close();

		generatePopulation( agents );
		printPopulation( agents, fname );
		
	}

	void generatePopulation( std::vector< std::shared_ptr< SimpleAgent > >& agents )
	{
		std::cout << "Database: " << database << std::endl;
		std::string fname( database + "TravelToWork.dat" );
		std::cout << "Preprocessing " << fname << " input file..." << std::endl;

		std::ifstream TTWFile( fname );
		if ( !TTWFile.good() )
		{
			std::cerr << "File " << fname << " does not exist." << std::endl;
		}

		/* Generate WG.dat file by counting how many people are in each DZ and dividing by 10 */

		/* map from DZ to WG ID to WG structure (holding number in group and number assigned to group) */
		std::map< dz_type, unsigned long > dz_numAgents_map;
		std::map< sla_type, unsigned long > sla_numStudents_map;

		// CZ: map SLA to number of people in each age group for validation
		std::vector<sla_type> SLA_list;
		std::map< sla_type, unsigned long > sla_numBabies_map;
		std::map< sla_type, unsigned long > sla_numChildren_map;
		std::map< sla_type, unsigned long > sla_numYoungAdults_map;
		std::map< sla_type, unsigned long > sla_numMiddleAged_map;
		std::map< sla_type, unsigned long > sla_numElderly_map;

		std::map< dz_type, std::vector< unsigned long > > dz_to_adult_idx;

		cd_dz_nagents_type cd_dz_nagents_map;

		unsigned long in_DZ, in_CD;
		unsigned int in_numAgents;
		while ( TTWFile >> in_CD >> in_DZ >> in_numAgents )
		{
			/* Initialise this entry if needed destination zone map */
			if ( dz_numAgents_map.find( in_DZ ) == dz_numAgents_map.end()  )
			{
				dz_numAgents_map[in_DZ] = 0;
			}

			dz_numAgents_map[in_DZ] += in_numAgents;
			cd_dz_nagents_map[in_CD].push_back( std::pair< dz_type, unsigned long >( in_DZ, in_numAgents ) );
		}
		TTWFile.close();

		/* Map all CDs to an SLA */
		std::map< cd_type, sla_type > cd_sla_map;
		std::ifstream cd_sla( database + "CDinSLA.dat" );
		unsigned long in_SLA;
		//float in_lat, in_long;
		while ( cd_sla >> in_CD >> in_SLA )//>> in_lat >> in_long )
		{
			cd_sla_map[in_CD] = in_SLA;
			sla_numStudents_map[in_SLA] = 0;
			//validation maps
			sla_numBabies_map[in_SLA] = 0;
			sla_numChildren_map[in_SLA] = 0;
			sla_numYoungAdults_map[in_SLA] = 0;
			sla_numMiddleAged_map[in_SLA] = 0;
			sla_numElderly_map[in_SLA] = 0;

			if (std::find(SLA_list.begin(), SLA_list.end(), in_SLA) == SLA_list.end())
			{
				SLA_list.push_back(in_SLA);
			}


		}
		cd_sla.close();

		/* How many working groups total? */
		unsigned long cumNumWGs = 0, numWGs;
		std::string wg_fname = output_directory + "WG.dat";
		std::ofstream WGFile( wg_fname );
		std::map< dz_type, struct WGLimitStruct > dz_wglim_map;
		struct WGLimitStruct wg_lim;
		for ( auto dz_to_numWg : dz_numAgents_map )
		{
			numWGs = ceil( (float)dz_to_numWg.second / (float)NUM_PEOPLE_IN_WG );
			WGFile << dz_to_numWg.first << " " << cumNumWGs << " " << numWGs << std::endl;
			wg_lim.lower = cumNumWGs;
			cumNumWGs += numWGs;
			wg_lim.upper = cumNumWGs;
			dz_wglim_map[dz_to_numWg.first] = wg_lim;
		}
		WGFile.close();

		std::cout << "Created " << cumNumWGs << " working groups in " << wg_fname << "." << std::endl;

		/* Create all working groups */
		int HH_count;

		std::cout << "Pre-loading family, age, house, school datafiles..." << std::endl;

		unsigned long min_CD = ULONG_MAX;
		unsigned long max_CD = 0;
		std::string cline;

		std::ifstream hh_comp_fs( database + "HOUSEHOLD_COMPOSITION.dat" );
		std::map< cd_type, std::vector< float > > hh_comp_map;
		{
			/* pre-load maps of CDs to data */
			float hh_comp_dist;
			while ( std::getline( hh_comp_fs, cline ) )
			{
				std::istringstream iss( cline );
				iss >> in_CD;
				hh_comp_map.emplace( in_CD, std::vector< float >() );
				while ( iss >> hh_comp_dist )
				{
					hh_comp_map.at( in_CD ).push_back( hh_comp_dist ); // building raw hh_comp distributions
				}
			}
		}
		hh_comp_fs.close();

		std::cout << "Household composition map size: " << hh_comp_map.size() << std::endl;

		auto hh_comp_it = hh_comp_map.begin();

		std::ifstream age_group_fs( database + "AGE.dat" );
		std::map< cd_type, std::pair< unsigned long, std::vector< float > > > age_map;
		{
			/* Load all age distributions */
			float age_dist;
			unsigned long cd_size;
			while ( std::getline( age_group_fs, cline ) )
			{
				std::istringstream iss( cline );
				iss >> in_CD >> cd_size;
				std::vector<float> dist;
				while ( iss >> age_dist )
				{
					dist.push_back( age_dist );
				}
				age_map.emplace( in_CD, std::pair< unsigned long, std::vector< float > >( cd_size, dist ) );
			}
		}
		age_group_fs.close();

		std::cout << "Agegroup map size: " << age_map.size() << std::endl;

		std::vector< unsigned > students;

		typedef std::vector< std::pair< AgeGroup, float > > agegroup_cdf_type;

		/* Generate the distributions */
		unsigned long n_CDs = 0;
		unsigned long agent_id = 0, hh_id = 0, hc_id = 0;
		unsigned long cum_agents = 0, n_adults = 0, n_children = 0;
		unsigned long n_agents_age[5] = {0};
		for ( auto age_it : age_map )
		{
			unsigned long c_CD = age_it.first;

			/* Check this CD is known to all data (census) files */
			if ( cd_sla_map.find( c_CD ) == cd_sla_map.end() )
			{
				std::cerr << std::endl << "CD " << c_CD << " not found in CDinSLA.dat." << std::endl;
				exit(1);
			}
			sla_type c_SLA = cd_sla_map.at( c_CD );

			if ( ( hh_comp_it = hh_comp_map.find( c_CD ) ) == hh_comp_map.end() )
			{
				std::cerr << std::endl << "CD " << c_CD << " not found in HOUSEHOLD_COMPOSITION_ABS.dat" << std::endl;
				continue;
			}

			unsigned long c_cd_size = age_it.second.first;
			auto& agegroup_dist = age_it.second.second;

			auto& hh_comp_dist = hh_comp_it->second;

			std::cout << "\rCreating CD:" << c_CD;
			//std::cout << "Creating CD:" << c_CD << std::endl;

			auto n_adult_males = agegroup_dist[SexAgeGroups::M19to29] + agegroup_dist[SexAgeGroups::M30to64] + agegroup_dist[SexAgeGroups::M65Plus];
			auto n_adult_females = agegroup_dist[SexAgeGroups::F19to29] + agegroup_dist[SexAgeGroups::F30to64] + agegroup_dist[SexAgeGroups::F65Plus];

			auto n_child_males = agegroup_dist[SexAgeGroups::M0to4] + agegroup_dist[SexAgeGroups::M5to18];
			auto n_child_females = agegroup_dist[SexAgeGroups::F0to4] + agegroup_dist[SexAgeGroups::F5to18];

			/*
			* Generate the CDFs for this census collection district.
			* generateCDF takes in the domain (first argument) and distribution (second argument) and returns a vector
			* to be used with sampleCDF (which returns a value from the domain).
			*/
			auto hh_comp_cdf = generateCDF( hh_comps, hh_comp_dist );

			auto adult_sex_cdf = generateCDF( sexes, std::vector< float >{ n_adult_males, n_adult_females } );

			auto child_sex_cdf = generateCDF( sexes, std::vector< float >{ n_child_males, n_child_females } );

			agegroup_cdf_type adult_male_agegroup_cdf = generateCDF( adult_agegroups, std::vector< float >{ agegroup_dist[SexAgeGroups::M19to29], agegroup_dist[SexAgeGroups::M30to64], agegroup_dist[SexAgeGroups::M65Plus] } );
			agegroup_cdf_type adult_female_agegroup_cdf = generateCDF( adult_agegroups,std::vector< float >{ agegroup_dist[SexAgeGroups::F19to29],agegroup_dist[SexAgeGroups::F30to64],agegroup_dist[SexAgeGroups::F65Plus] } );

			agegroup_cdf_type child_male_agegroup_cdf = generateCDF( child_agegroups,std::vector< float >{ agegroup_dist[SexAgeGroups::M0to4],agegroup_dist[SexAgeGroups::M5to18] } );
			agegroup_cdf_type child_female_agegroup_cdf = generateCDF( child_agegroups,std::vector< float >{ agegroup_dist[SexAgeGroups::F0to4],agegroup_dist[SexAgeGroups::F5to18] } );

			agegroup_cdf_type adult_at_parents_agegroup_cdf = generateCDF( std::vector< AgeGroup >{ AgeGroups::Age19to29 },std::vector< float >{ 1.0 } );

			unsigned long cum_agents_in_cd = 0;
			unsigned long n_hh_in_cluster = 0;
			while ( cum_agents_in_cd < c_cd_size )
			{
				/* Draw the current household composition */
				/*
				* TODO: make sure composition size is less than/equal to # agents left in CD
				*/
				HouseholdComposition c_hh_comp = HouseholdComposition( sampleCDF( hh_comp_cdf ) );
				//hh_comp_drawn[c_hh_comp]++;

				/* Vector of CDFs that we need to sample to get age groups of the agents */
				std::vector< agegroup_cdf_type > agegroup_cdfs;
				unsigned long n_adults_to_sample = 0;
				unsigned long n_children_to_sample = 0;

				/* Single/lone household compositions */
				if ( c_hh_comp <= HouseholdCompositions::Single )
				{
					n_adults_to_sample = 1;
				}
				/* Couple with/without children household compositions */
				else if ( c_hh_comp <= HouseholdCompositions::CouplesWithFourChildren )
				{
					agegroup_cdfs.push_back( adult_male_agegroup_cdf );
					agegroup_cdfs.push_back( adult_female_agegroup_cdf );
					n_adults += 2;

					/* Couple with children household compositions */
					if ( c_hh_comp >= HouseholdCompositions::CouplesWithOneChild )
					{
						n_children_to_sample = c_hh_comp - HouseholdCompositions::CouplesWithOneChild + 1;
					}
				}
				/* Single parent family household compositions */
				else if ( c_hh_comp <= HouseholdCompositions::SingleParentFamilyFiveChildren )
				{
					n_adults_to_sample = 1;
					n_children_to_sample = c_hh_comp - HouseholdCompositions::SingleParentFamilyOneChild + 1;
				}
				/* Group household compositions */
				else
				{
					n_adults_to_sample = c_hh_comp - HouseholdCompositions::GroupTwo + 2;
				}

				unsigned n_adults_at_parents = 0;
				for ( unsigned child_id = 0; child_id < n_children_to_sample; ++child_id )
				{
					/* Are these "children" actually adults living at home? */
					if ( rand() / float( RAND_MAX ) <= PERC_ADULTS_AT_HOME )
					{
						agegroup_cdfs.push_back( adult_at_parents_agegroup_cdf );
						n_adults_at_parents++;
					}
					/* Else add the children (<18) age group CDFs for sampling */
					else if ( sampleCDF( child_sex_cdf ) == Sex::Male )
					{
						agegroup_cdfs.push_back( child_male_agegroup_cdf );
					}
					else
					{
						agegroup_cdfs.push_back( child_female_agegroup_cdf );
					}
				}
				n_children += n_children_to_sample - n_adults_at_parents;


				for ( unsigned adult_id = 0; adult_id < n_adults_to_sample; ++adult_id )
				{
					if ( sampleCDF( adult_sex_cdf ) == Sex::Male )
					{
						agegroup_cdfs.push_back( adult_male_agegroup_cdf );
					}
					else
					{
						agegroup_cdfs.push_back( adult_female_agegroup_cdf );
					}
				}
				n_adults += n_adults_to_sample + n_adults_at_parents;

				for ( auto c_cdf : agegroup_cdfs )
				{
					auto cagent = generateAgent( agent_id, c_CD, hh_id, hc_id, c_cdf );



					agents.push_back( cagent );

					n_agents_age[cagent->my_age]++;
					// CZ: should update the numStudents_sla_map here, not in the WG conditional
					if ( cagent->isStudentAge() )
					{
						if ( rand() / float(RAND_MAX) < SCHOOL_ATTENDANCE_LIKELIHOOD )
						{
							students.push_back( agent_id );
							if ( sla_numStudents_map.find( c_SLA ) == sla_numStudents_map.end() )
							{
								sla_numStudents_map[c_SLA] = 0;
							}
							sla_numStudents_map.at(c_SLA)++;
						}
					}

					agent_id++ ;

					// add agents to validation maps
					if (cagent-> isBaby())
					{
						if (sla_numBabies_map.find(c_SLA) == sla_numBabies_map.end())
						{sla_numBabies_map[c_SLA] = 0;}
						sla_numBabies_map.at(c_SLA)++;
					}

					if (cagent-> isStudentAge())
					{
						if (sla_numChildren_map.find(c_SLA) == sla_numChildren_map.end())
						{sla_numChildren_map[c_SLA] = 0;}
						sla_numChildren_map.at(c_SLA)++;
					}

					if (cagent-> isYoungAdult())
					{
						if (sla_numYoungAdults_map.find(c_SLA) == sla_numYoungAdults_map.end())
						{sla_numYoungAdults_map[c_SLA] = 0;}
						sla_numYoungAdults_map.at(c_SLA)++;
					}

					if (cagent-> isMiddleAged())
					{
						if (sla_numMiddleAged_map.find(c_SLA) == sla_numMiddleAged_map.end())
						{sla_numMiddleAged_map[c_SLA] = 0;}
						sla_numMiddleAged_map.at(c_SLA)++;
					}

					if (cagent-> isElderly())
					{
						if (sla_numElderly_map.find(c_SLA) == sla_numElderly_map.end())
						{sla_numElderly_map[c_SLA] = 0;}
						sla_numElderly_map.at(c_SLA)++;
					}

				}

				hh_id++;
				/* Ensure we don't have clusters going between CDs */
				if ( ++n_hh_in_cluster == 4 )
				{
					hc_id++;
					n_hh_in_cluster = 0;
				}

				cum_agents_in_cd += agegroup_cdfs.size();
			}
			hc_id++;

			cd_dz_nagents_type::iterator cd_dz_nagents_it;

			std::map< dz_type, std::map< wg_type, unsigned long > > dz_wg_num_map;

			/* If there are actually adults commuting from this CD to any DZ... */
			if ( ( cd_dz_nagents_it = cd_dz_nagents_map.find( c_CD ) ) != cd_dz_nagents_map.end() )
			{
				/* ...pre-select all adults in area. */
				std::vector< unsigned long > adult_idx_in_dz;
				for ( unsigned i = 0; i < cum_agents_in_cd; ++i )
				{
					unsigned long agent_id = cum_agents + i;
					if ( agents[agent_id]->isAdult() )
					{
						/* 35% of agents aged 65+ work */
						if ( agents[agent_id]->my_age < AgeGroups::Age65Plus ||
							rand() / float(RAND_MAX) < 0.35 )
							{
								adult_idx_in_dz.push_back( agent_id );

								if (agents[agent_id]->isStudentAge())
								{
									std::cout<< "error: assigned a student to a working group" << std::endl;

									//std::cout << "Press enter to continue ...";
									//std::cin.get();
								}

							}
						}
						// CZ:: why is this check conditional on commuters leaving the CD?
						/*else if ( agents[agent_id]->isStudentAge() )
						{
						if ( rand() / float(RAND_MAX) < SCHOOL_ATTENDANCE_LIKELIHOOD )
						{
						students.push_back( agent_id );
						if ( sla_numStudents_map.find( c_SLA ) == sla_numStudents_map.end() )
						{
						sla_numStudents_map[c_SLA] = 0;
					}
					sla_numStudents_map.at(c_SLA)++;
				}
			}*/
		}

		/* Iterate through each DZ this CD can go to. */
		for ( auto dz_to_nagents_it : cd_dz_nagents_it->second )
		{
			dz_type c_DZ = dz_to_nagents_it.first;
			unsigned long n_commuting_agents = dz_to_nagents_it.second;

			if ( dz_to_adult_idx.find( c_DZ ) == dz_to_adult_idx.end() )
			{
				dz_to_adult_idx[ c_DZ ] = std::vector< unsigned long >();
			}

			/* Gives which working groups we can select from */
			struct WGLimitStruct wg_lims = dz_wglim_map.at( c_DZ );
			unsigned wg_range = wg_lims.upper - wg_lims.lower;

			if ( dz_wg_num_map.find( c_DZ ) == dz_wg_num_map.end() )
			{
				dz_wg_num_map[ c_DZ ] = std::map< wg_type, unsigned long >();
			}

			/*
			* Now, randomly assign each agent from list to a working group
			*(don't care how many are in each group because it should be E{18} due to number of groups)
			*/
			for ( unsigned i = 0; i < n_commuting_agents; ++i )
			{
				if ( adult_idx_in_dz.size() == 0 )
				{
					std::cerr << std::endl << "Number of commuting agents (" << n_commuting_agents - i
					<< " more commuters than working adults in DZ:" << c_DZ << ")." << std::endl;
					break;
				}
				unsigned long selectedAdult = rand() % adult_idx_in_dz.size();//locates a random index
				wg_type selectedWG = rand() % wg_range + wg_lims.lower;//selects a random working group index

				if ( selectedWG > wg_lims.upper )
				{
					std::cerr << "err" << std::endl;
				}

				dz_to_adult_idx[c_DZ].push_back( adult_idx_in_dz.at( selectedAdult ) ); // adds agent to dz->worker map

				agents[adult_idx_in_dz.at( selectedAdult )]->my_wg = selectedWG;

				if (agents[adult_idx_in_dz.at(selectedAdult)]->isChild())
				{
					std::cout<<"error: assigned a child to a working group"<<std::endl;
				}


				adult_idx_in_dz.erase( adult_idx_in_dz.begin() + selectedAdult );

				if ( dz_wg_num_map.at( c_DZ ).find( c_DZ ) == dz_wg_num_map.at( c_DZ ).end() )
				{
					dz_wg_num_map.at( c_DZ )[selectedWG] = 0;
				}
				dz_wg_num_map.at( c_DZ )[ selectedWG ]++;
			}
		}

	}

	cum_agents += cum_agents_in_cd;
}

std::cout << std::endl << "Created " << agents.size() << " agents"
<< " comprising " << n_adults << " adults and " << n_children << " children. "
<<  students.size() << " of which are students" << std::endl;

std::cout << "Agent in age groups: ";
for ( unsigned i = 0; i < 5; ++i )
{
	std::cout << n_agents_age[i] << " ";
}
std::cout << std::endl;

// output an SLA to age-group file for validation purposes

std::cout<< "printing validation data..." << std::endl;

std::ofstream SLA_to_AgeGroups( output_directory + "SLA_Age_output.dat");

for (int i = 0; i < SLA_list.size(); i++)
{
	in_SLA = SLA_list[i];
	SLA_to_AgeGroups << in_SLA << " " << sla_numBabies_map[in_SLA] << " "
	<< sla_numChildren_map[in_SLA] << " " << sla_numYoungAdults_map[in_SLA] << " "
	<< sla_numMiddleAged_map[in_SLA] << " " << sla_numElderly_map[in_SLA] << std::endl;
}

SLA_to_AgeGroups.close();


//std::cout << "Press enter to continue ...";
//std::cin.get();

/*
* Make schools and assign teachers and students
*/

std::cout << "generating school configuration" << std::endl;

cd_dz_nagents_map_type student_adjacency_list;

// make a stream to write which DZ schools are in
std::ofstream school_in_dz( output_directory + "SCHOOLinDZ.dat" );

// make a map of SLAs to the schools they contain
sla_school_map_type sla_school_map; // DECLARATION OF FUNCTION INPUT FOR SCHOOL_CONFIGURATION()

//make a map of school ids to school objects
school_id_map_type school_id_map; // DECLARATION OF FUNCTION INPUT FOR SCHOOL_CONFIGURATION()

//map SLA to vector of DZs
sla_dz_map_type sla_dz_map; // DECLARATION OF FUNCTION INPUT FOR SCHOOL_CONFIGURATION()

std::vector< sla_type > sla_list;
std::ifstream SLA_to_DZ_file(database + "DZinSLA.dat");

while (SLA_to_DZ_file >> in_SLA >> in_DZ)
{
	if (sla_dz_map.find(in_SLA) == sla_dz_map.end())
	{
		sla_dz_map[in_SLA] = std::vector< dz_type >();
		sla_school_map[in_SLA] = std::vector< struct School* >();
		sla_list.push_back(in_SLA);
	}
	sla_dz_map[in_SLA].push_back(in_DZ);
}
SLA_to_DZ_file.close();
/*for (size_t i = 0; i < sla_list.size(); i++ )
{
std::cout << "SLA " << sla_list[i] << " contains DZs: " << std::endl;
for (size_t j = 0; j < sla_dz_map[sla_list[i]].size(); j++)
{
std::cout<< "\t" << sla_dz_map[sla_list[i]][j] << std::endl;
}
}*/

std::cout << "Number of SLAs with DZs: " << sla_school_map.size() << std::endl;


// map dz to number of workers
// make dz_numAgents_map from dz_to_adult_idx[DZ].size()
dz_numAgents_map_type dz_numWorkers_map; // DECLARATION OF FUNCTION INPUT FOR SCHOOL_CONFIGURATION()
dz_type DZ_i;
std::map< dz_type, std::vector< unsigned long > >::iterator dz_to_adult_idx_it = dz_to_adult_idx.begin();

for(dz_to_adult_idx_it; dz_to_adult_idx_it != dz_to_adult_idx.end(); ++dz_to_adult_idx_it)
{
	DZ_i = dz_to_adult_idx_it->first;
	dz_numWorkers_map[DZ_i] = dz_to_adult_idx[DZ_i].size();
	//std::cout << "DZ " << DZ_i << " contains " << dz_numWorkers_map[DZ_i] << " workers" << std::endl;
}


std::cout << "opening postcode_to_SLA correspondence list" << std::endl;


//map postcodes to vectors of SLAs
postcode_sla_map_type postcode_sla_map; // DECLARATION OF FUNCTION INPUT FOR SCHOOL CONFIGURATION()
std::vector< postcode_type > postcode_list;
std::ifstream PC_to_SLA_file(database + "Postcode_SA2_2021.dat");
unsigned long in_PC;
while (PC_to_SLA_file >> in_PC >> in_SLA)
{
	if (postcode_sla_map.find(in_PC) == postcode_sla_map.end())
	{
		postcode_sla_map[in_PC] = std::vector< sla_type >();
		postcode_list.push_back(in_PC);
	}
	postcode_sla_map[in_PC].push_back(in_SLA);
}
PC_to_SLA_file.close();
/*for (size_t i = 0; i < postcode_list.size(); i++ )
{
std::cout << "Postcode " << postcode_list[i] << " contains SLAs: " << std::endl;
for (size_t j = 0; j < postcode_sla_map[postcode_list[i]].size(); j++)
{
std::cout<< "\t" << postcode_sla_map[postcode_list[i]][j] << std::endl;
}
}*/


// map school to postcode and school size
school_postcode_map_type school_postcode_map; // DECLARATION OF FUNCTION INPUT FOR SCHOOL CONFIGURATION()
school_schoolsize_map_type school_schoolsize_map; // DECLARATION OF FUNCTION INPUT FOR SCHOOL CONFIGURATION()
schoolsize_school_map_type schoolsize_school_map; // DECLARATION OF FUNCTION INPUT FOR SCHOOL CONFIGURATION()
std::vector< unsigned long > school_list;
std::ifstream school_PC_file(database + "postcode_enrolment_2021.dat");
unsigned long in_school;
in_school = 0;
unsigned long in_school_size;
while ( school_PC_file >> in_PC >> in_school_size )
{
	if (school_postcode_map.find(in_school) == school_postcode_map.end())
	{
		school_postcode_map[in_school] = 0;
		school_list.push_back(in_school);
	}
	school_postcode_map[in_school] = in_PC;

	if (school_schoolsize_map.find(in_school) == school_schoolsize_map.end())
	{
		school_schoolsize_map[in_school] = 0;
	}
	school_schoolsize_map[in_school] = in_school_size;

	if (schoolsize_school_map.find(in_school_size) == schoolsize_school_map.end())
	{
		schoolsize_school_map[in_school_size] = std::vector<unsigned long>();
	}
	schoolsize_school_map[in_school_size].push_back(in_school);

	in_school++;
}
school_PC_file.close();

//check the schoolsize_school_map here, something may be wrong with it.
/*schoolsize_school_map_type::iterator size_school_map_it = schoolsize_school_map.begin();

for(size_school_map_it; size_school_map_it!=schoolsize_school_map.end();++size_school_map_it)
{
std::cout<<" school(s)";
for(int i = 0; i < size_school_map_it->second.size(); i++)
{
std::cout<< ", " << size_school_map_it->second[i];
}
std::cout<< " are of size " << size_school_map_it->first << std::endl;
}

std::cout << "Press enter to continue ...";
std::cin.get();*/


/*			for (size_t i = 0; i < school_list.size(); i++ )
{
std::cout << "School " << school_list[i]<< " of size "
<< school_schoolsize_map[school_list[i]] << " is in postcode "
<< school_postcode_map[school_list[i]] << std::endl;
}*/



// map number of student-aged people to vector of SLAs with that number of student-aged people
// use existing map sla_numStudents_map
numStudents_sla_map_type numStudents_sla_map; // DECLARATION OF FUNCTION INPUT FOR SCHOOL CONFIGURATION()
unsigned long in_nStudents;
std::map< sla_type, unsigned long >::iterator sla_numStudents_map_it = sla_numStudents_map.begin();
for (sla_numStudents_map_it; sla_numStudents_map_it!=sla_numStudents_map.end(); ++sla_numStudents_map_it)
{
	in_SLA = sla_numStudents_map_it->first;
	in_nStudents = sla_numStudents_map_it->second;

	//std::cout << "SLA " << in_SLA << " contains " << in_nStudents << " students" << std::endl;

	if (numStudents_sla_map.find(in_nStudents) == numStudents_sla_map.end())
	{
		numStudents_sla_map[in_nStudents] = std::vector<sla_type>();
	}
	numStudents_sla_map[in_nStudents].push_back(in_SLA);
}



school_configuration( schoolsize_school_map, school_postcode_map, school_schoolsize_map,
	postcode_sla_map, sla_numStudents_map, sla_dz_map, dz_numWorkers_map,
	numStudents_sla_map, sla_school_map, school_id_map);

	// new school configuration code here

	//school_in_dz.close();

	/*
	* Load pairwise distances between SLAs
	*/
	std::map< sla_type, std::map< float, sla_type > > sla_distance_sla_map;
	std::ifstream sla_distance_matrix( database + "DistanceSLA.txt" );
	std::cout << "Loading SLA distance matrix" << std::endl;

	/* First get which SLAs we have distances for... */
	std::getline( sla_distance_matrix, cline );
	std::istringstream iss( cline );
	sla_type csla;
	std::vector< sla_type > sla_list_distmat;
	while ( iss >> csla )
	{
		sla_list_distmat.push_back( csla );
	}
	std::cout << "Stored " << sla_list_distmat.size() << " SLAs" << std::endl;

	/* ...then for every SLA, create a map of (every other) SLA to the distance between them */
	unsigned sla_row_idx = 0;
	while ( std::getline( sla_distance_matrix, cline ) )
	{
		std::istringstream iss( cline );
		float dist;
		sla_type sla_row = sla_list_distmat.at( sla_row_idx++ );
		sla_distance_sla_map[ sla_row ].emplace();

		for ( sla_type sla_col_idx = 0; iss >> dist; ++sla_col_idx )
		{
			if ( dist < MAX_STUDENT_COMMUTE_DISTANCE )
			{
				sla_type sla_col = sla_list_distmat.at( sla_col_idx );
				sla_distance_sla_map.at( sla_row )[ dist ] = sla_col;
			}
		}
	}

	std::cout << "done interpreting distance matrix" << std::endl;

	unsigned num_without_schools = 0;

	/* Generate CDFs for every entry in the SLA distance matrix */
	std::map< sla_type, std::map< float, std::vector< std::pair< unsigned long, float > > > > sla_dist_school_id_cdf_map;
	std::map< sla_type, float > school_zone_scaler;
	std::cout << "Generating CDFs for all SLA->SLA distances..." << std::endl;
	unsigned long i = 0;
	for ( const auto& distance_sla_map : sla_distance_sla_map )
	{
		auto from_SLA = distance_sla_map.first;
		sla_dist_school_id_cdf_map[ from_SLA ].emplace();

		unsigned long total_school_sizes = 0;
		std::vector< unsigned long > school_ids;
		std::vector< unsigned long > school_sizes;

		for ( const auto& centry : distance_sla_map.second )
		{
			auto cdist = centry.first;
			auto to_SLA = centry.second;

			/* This is an ordered map, so every distance entry before is less than this distance */
			if ( sla_school_map.find( to_SLA ) != sla_school_map.end() )
			{
				for ( const auto& cschool : sla_school_map.at( to_SLA ) )
				{
					school_ids.push_back( cschool->ID );
					school_sizes.push_back( cschool->max_students );
				}
			}

			if ( school_sizes.size() > 0 )
			{
				std::vector< float > school_dist;
				/* Re-normalise distributions */
				for ( auto csize : school_sizes )
				{
					school_dist.push_back( (float)csize );
				}
				sla_dist_school_id_cdf_map.at( from_SLA )[ cdist ] = generateCDF( school_ids, school_dist );

				//std::cout << "CDF from SLA:" << from_SLA << " to SLA:" << to_SLA << " (distance: " << cdist
				//		<< ") has " << school_ids.size() << " schools in it." << std::endl;

			}
		}
		std::cout << "\rGenerated CDF for SLA:" << from_SLA
		<< " (number " << ++i << "/" << sla_distance_sla_map.size() << "; num distances: "
		<< sla_dist_school_id_cdf_map.at( from_SLA ).size() << "; num schools: "
		<< sla_dist_school_id_cdf_map.at( from_SLA ).rbegin()->second.size() << ")" << std::fflush;

		school_zone_scaler[from_SLA] = 1.0;

		if ( sla_dist_school_id_cdf_map.at( from_SLA ).rbegin()->second.size() == 0 )
		{
			num_without_schools++;
		}
	}
	std::cout << std::endl;

	std::cout << "Number of SLAs without schools within " << MAX_STUDENT_COMMUTE_DISTANCE * 100 << "km: " << num_without_schools << std::endl;

	unsigned long n_students_assigned = 0, n_students_assigned_to_their_sla = 0, n_students_not_assigned = 0;

	/* Assign students to schools based on zones */
	unsigned long tot_num_students = students.size();

	float tot_school_dist = 0.0;
	while ( students.size() > 0 )
	{
		unsigned selected_student = rand() % students.size();
		unsigned agent_idx = students.at( selected_student );
		students.erase( students.begin() + selected_student );

		if ( agents[agent_idx]->my_wg != 0 )
		{
			if (agents[agent_idx]->isAdult())
			{
				std::cout << "Student #:" <<  agent_idx << "is an adult" << std::endl;
			}
			std::cerr << "Student #:" << agent_idx << " already assigned WG." << std::endl;
			continue;

		}

		sla_type c_SLA = cd_sla_map.at( agents[agent_idx]->my_cd );

		if ( sla_distance_sla_map.find( c_SLA ) == sla_distance_sla_map.end() )
		{
			std::cerr << "Student #:" << agent_idx << " in unknown SLA:" << c_SLA << std::endl;
			// this occurs if the SLA is in the migratory-offshore-shipping category
			continue;
		}


		auto& dist_school_id_cdf_map = sla_dist_school_id_cdf_map.at( c_SLA );

		bool increase_zone = false;

		if ( dist_school_id_cdf_map.size() == 0 || dist_school_id_cdf_map.rbegin()->second.size() == 0 )
		{
			//std::cerr << "SLA: " << c_SLA << " has no schools within " << MAX_STUDENT_COMMUTE_DISTANCE * 100 << "km" << std::endl;
		}
		else
		{
			for ( unsigned i = 0; agents.at( agent_idx )->my_wg >= 0 && i < MAX_ALLOCATION_ATTEMPTS; ++i )
			{
				/* Use CDF to find  */
				float chosen_dist = invFoldedNormCDF( school_zone_scaler.at( c_SLA ) * SCHOOL_ZONE_SIGMA, rand() / float(RAND_MAX) );

				/* Find the closest SLA that is within this distance */
				auto farthest_dist_it = dist_school_id_cdf_map.begin();

				if ( dist_school_id_cdf_map.size() > 1 )
				{
					if ( ( farthest_dist_it = dist_school_id_cdf_map.lower_bound( chosen_dist ) ) == dist_school_id_cdf_map.end() )
					{
						/*
						std::cerr << "SLA: " << c_SLA << " has no schools within " << chosen_dist * 100 << "km. Max distance is "
						<< dist_school_id_cdf_map.rbegin()->first * 100 << "km"
						<< " (num distances: " << dist_school_id_cdf_map.size() << ")." << std::endl;
						*/
						continue;
					}
				}

				auto school_id_cdf = farthest_dist_it->second;

				unsigned long chosen_school_id = sampleCDF( school_id_cdf );
				/*
				std::cout << "Chose School:" << chosen_school_id << " for agent " << agent_idx
				<< " (allowable catchment: " << chosen_dist << "; num SLAs: "
				<< std::distance( dist_school_id_cdf_map.begin(), farthest_dist_it )
				<< "; num schools: "<< school_id_cdf.size() << ")." << std::endl;
				*/
				struct School * my_school = school_id_map.at( chosen_school_id );

				if ( my_school->num_students >= my_school->max_students )
				{

					/*std::cerr << "School:" << my_school->ID << " at capacity (" << school_id_cdf.size()
					<< " schools within chosen dist: " << chosen_dist
					<< "; schools within 150km of " << c_SLA << ":" << dist_school_id_cdf_map.rbegin()->second.size() << ")." << std::endl;
					*/
					increase_zone = true;
					continue;
				}

				agents[agent_idx]->my_wg = -my_school->ID;
				my_school->num_students++;


				//CZ: add element to student adjacency list

				std::pair<cd_type, dz_type> CD_DZ_student (agents[agent_idx]->my_cd, my_school->DZ);
				if (student_adjacency_list.find(CD_DZ_student) == student_adjacency_list.end() )
				{
					student_adjacency_list[CD_DZ_student] = 1;
				}
				else
				{
					student_adjacency_list[CD_DZ_student]++;
				}

				/*std::cout<<"updating student adjacency list. CD " << agents[agent_idx]->my_cd
				<< " to DZ " << my_school->DZ << " has "
				<< student_adjacency_list[CD_DZ_student] << " students" << std::endl;*/


				if ( my_school->SLA == c_SLA )
				{
					n_students_assigned_to_their_sla++;
				}
				n_students_assigned++;

				tot_school_dist += chosen_dist;
			}
		}

		if ( agents[agent_idx]->my_wg >= 0 )
		{
			n_students_not_assigned++;
			if ( increase_zone )
			{
				school_zone_scaler.at( c_SLA ) += 0.1;
			}
		}

		if ( ( n_students_assigned + n_students_not_assigned ) % 1000 == 0 )
		{
			std::cout << "\r# students assigned: " << n_students_assigned << "/" << tot_num_students
			<< " (" << students.size() << " students left). # assigned to their SLA: " << n_students_assigned_to_their_sla
			<< " (" << n_students_assigned - n_students_assigned_to_their_sla << " commuters; "
			<< n_students_not_assigned << " not assigned; avg dist: " << tot_school_dist / (float)n_students_assigned << ")." << std::flush;
		}
	}

	std::cout << "# students assigned: " << n_students_assigned << ". # assigned to their SLA: " << n_students_assigned_to_their_sla
	<< "(" << n_students_assigned - n_students_assigned_to_their_sla << " commuters; " << n_students_not_assigned << "not assigned)." << std::endl;

	unsigned long n_teachers_assigned = 0;
	/* Iterate through all SLAs */
	for ( auto& sla_school_it : sla_school_map )
	{
		/* Iterate through all schools in this SLA */

		for ( auto& cschool : sla_school_it.second )
		{
			dz_type currentDZ = cschool->DZ;

			unsigned num_teachers_required = ceil( (float)cschool->num_students / (float)NUM_STUDENTS_TO_TEACHERS );

			std::cout << "\rNumber of students in school: " << cschool->ID << " is " << cschool->num_students << std::endl;

			/* Randomly pick the teachers from this DZ */
			if ( dz_to_adult_idx.find( currentDZ ) == dz_to_adult_idx.end() )
			{
				std::cerr << "Could not find DZ:" << currentDZ << " in list." << std::endl;
				continue;
			}
			auto& potential_adults = dz_to_adult_idx.at( currentDZ );

			unsigned num_teachers_assigned = 0;

			while ( num_teachers_assigned < num_teachers_required )
			{

				std::cout << "current DZ " << currentDZ << " has potential adults: " << potential_adults.size() << std::endl;
				unsigned long rval = rand() % potential_adults.size(); // error if not enough teachers

				unsigned long adult_idx = potential_adults.at( rval );
				agents[adult_idx]->my_wg = -cschool->ID;
				num_teachers_assigned++;

				std::cout << "\rAssigned adult " << adult_idx << " (out of " << potential_adults.size()
				<< ") to school:" << cschool->ID
				<< " (" << num_teachers_assigned << "/" << num_teachers_required << ")";

				potential_adults.erase( potential_adults.begin() + rval );
				n_teachers_assigned++;

				if ( potential_adults.size() == 0 )
				{
					std::cerr << "No potential adults left in DZ:" << currentDZ << std::endl;
					break;
				}
			}
			//CZ moved this from immediately after school initialization
			//school_in_dz << cschool->ID << " " << cschool->DZ << " " << cschool->max_students << " " << cschool->num_student << std::endl;
			school_in_dz << cschool->ID << " " << cschool->DZ << " " << cschool->num_students << std::endl;
		}
	}
	std::cout << "Number of teachers assigned: " << n_teachers_assigned << std::endl;
	school_in_dz.close();


//CZ: print out the student adjacency list
// make output file stream
std::cout << "writing student adjacency list..."<< std::endl;

std::ofstream CD_DZ_nStudents( output_directory + "CD_DZ_Nstudents.dat" );
cd_dz_nagents_map_type::iterator student_adjacency_list_it = student_adjacency_list.begin();
for (student_adjacency_list_it; student_adjacency_list_it!=student_adjacency_list.end(); ++student_adjacency_list_it)
{
CD_DZ_nStudents << student_adjacency_list_it->first.first << " "
 << student_adjacency_list_it->first.second << " "
 << student_adjacency_list_it->second << std::endl;
}
// close output file stream
CD_DZ_nStudents.close();
std::cout << "Done."<< std::endl;
}

std::shared_ptr< SimpleAgent > generateAgent( unsigned long id, cd_type cd, unsigned long hh, unsigned long hc,
	const std::vector< std::pair< AgeGroup, float > >& age_cdf )
	{
		std::shared_ptr< SimpleAgent > nagent = std::make_shared< SimpleAgent >();
		nagent->my_id = id;
		nagent->my_hh = hh;
		nagent->my_hc = hc;
		nagent->my_cd = cd;
		nagent->my_wg = 0;

		/* Sample my age from the age CDF */
		nagent->my_age = sampleCDF( age_cdf );
		return nagent;
	}

	void printPopulation( const std::vector< std::shared_ptr< SimpleAgent > >& agents, const std::string& fname )
	{
		std::ofstream pFile( fname );
		for ( auto& cagent : agents )
		{
			pFile << cagent->my_age << " " << cagent->my_hh << " " << cagent->my_hc << " " << cagent->my_wg << " " << cagent->my_cd << std::endl;
		}
		pFile.close();
	}

	void school_configuration(
		schoolsize_school_map_type& schoolsize_school_map,
		school_postcode_map_type& school_postcode_map,
		school_schoolsize_map_type& school_schoolsize_map,
		postcode_sla_map_type& postcode_sla_map,
		sla_numStudents_map_type& sla_numStudents_map,
		sla_dz_map_type& sla_dz_map,
		dz_numAgents_map_type& dz_numWorkers_map,
		numStudents_sla_map_type& numStudents_sla_map,
		sla_school_map_type& sla_school_map,
		school_id_map_type& school_id_map
	)
	{

		// write some input checks here, something is haywire

		// check DZ to SLA map
		/*sla_dz_map_type::iterator sla_dz_test_it = sla_dz_map.begin();
		for (sla_dz_test_it; sla_dz_test_it!=sla_dz_map.end(); ++sla_dz_test_it)
		{
		std::cout << "SLA " << sla_dz_test_it->first << " contains DZs: ";
		for (int i = 0; i < sla_dz_test_it->second.size(); i++)
		{
		std::cout << sla_dz_test_it->second[i] << " ";
	}
	std::cout<<std::endl;
}
*/

// check schoolsize_school_map
/*schoolsize_school_map_type::iterator test_it = schoolsize_school_map.begin();
for (test_it; test_it!=schoolsize_school_map.end(); ++test_it)
{
std::cout << "size " << test_it->first << " applies to schools: ";
for (int i = 0; i < test_it->second.size(); i++)
{
std::cout << test_it->second[i] << " ";
}
std::cout<<std::endl;
}*/

//check school_postcode_map
/*school_postcode_map_type::iterator test_it1 = school_postcode_map.begin();
for (test_it1; test_it1!=school_postcode_map.end(); ++test_it1)
{
std::cout << "school " << test_it1->first << " is in postcode: "<< test_it1->second <<std::endl;
}
std::cout << "Press enter to continue ...";
std::cin.get();
*/

//check school->schoolsize maps
/*school_schoolsize_map_type::iterator test_it2 = school_schoolsize_map.begin();
for (test_it2; test_it2!=school_schoolsize_map.end(); ++test_it2)
{
std::cout << "school " << test_it2->first << " is of size: "<< test_it2->second <<std::endl;
}
std::cout << "Press enter to continue ...";
std::cin.get();
*/

//check postcode->sla map
/*postcode_sla_map_type::iterator test_it3 = postcode_sla_map.begin();
for (test_it3; test_it3!=postcode_sla_map.end(); ++test_it3)
{
std::cout << "postcode " << test_it3->first << " contains slas: ";
for (int i = 0; i < test_it3->second.size(); i++)
{
std::cout << test_it3->second[i] << " ";
}
std::cout<<std::endl;
}
std::cout << "Press enter to continue ...";
std::cin.get();
*/

//check sla->numstudents map
/*sla_numStudents_map_type::iterator test_it4 = sla_numStudents_map.begin();
for (test_it4; test_it4!=sla_numStudents_map.end(); ++test_it4)
{
std::cout << "sla " << test_it4->first << " contains: "<< test_it4->second << " student-aged people" << std::endl;
}
std::cout << "Press enter to continue ...";
std::cin.get();
*/

//check dz->numWorkers map
/*dz_numAgents_map_type::iterator test_it5 = dz_numWorkers_map.begin();
for (test_it5; test_it5!=dz_numWorkers_map.end(); ++test_it5)
{
std::cout << "dz " << test_it5->first << " contains: "<< test_it5->second << " workers" << std::endl;
}
std::cout << "Press enter to continue ...";
std::cin.get();
*/

//check numstudents->sla map
/*numStudents_sla_map_type::iterator test_it6 = numStudents_sla_map.begin();
for (test_it6; test_it6!=numStudents_sla_map.end(); ++test_it6)
{
std::cout << "nstudents " << test_it6->first << " applies to SLAs: ";
for (int i = 0; i < test_it6->second.size(); i++)
{
std::cout << test_it6->second[i] << " ";
}
std::cout<<std::endl;
}
std::cout << "Press enter to continue ...";
std::cin.get();
*/

std::ofstream ACARA_Discrepency(output_directory + "ACARA_Discrepency.dat", std::ios::app);

schoolsize_school_map_type::reverse_iterator size_school_map_it = schoolsize_school_map.rbegin();
std::vector<unsigned long> schools_of_size_x;
unsigned long postcode_school_x;
std::vector<sla_type> SLAs_in_postcode_x;
unsigned long local_student_population;
sla_type SLA_school_x;

std::vector<dz_type> DZs_in_SLA_x;
unsigned long num_teachers_required;
dz_type DZ_school_x;

for (size_school_map_it; size_school_map_it!= schoolsize_school_map.rend(); ++size_school_map_it)
{
	schools_of_size_x = size_school_map_it->second;

	while (schools_of_size_x.size()!=0)
	{
		//find postcode
		postcode_school_x = school_postcode_map[schools_of_size_x.front()];
		//find SLA with most student-aged individuals
		SLAs_in_postcode_x = postcode_sla_map[postcode_school_x];
		local_student_population = 0;

		std::cout<< "school " << schools_of_size_x.front() << " is in postcode " << postcode_school_x
		<< ", and is size " << size_school_map_it->first << std::endl;

		bool SLA_assignment_flag = false;
		bool random_SLA_flag = false;
		// this will be switched if postcode runs out of students,
		// in this case a random SLA with enough possible workers in some DZ will be selected

		std::cout << "Postcode " << postcode_school_x << " contains " << SLAs_in_postcode_x.size() << " SLAs" << std::endl;

		for (int i = 0; i<SLAs_in_postcode_x.size(); i++)
		{
			std::cout << "SLA " << SLAs_in_postcode_x[i] << " contains " <<
			sla_numStudents_map[SLAs_in_postcode_x[i]] << " unallocated students" << std::endl;

			if (sla_numStudents_map[SLAs_in_postcode_x[i]] > local_student_population)
			{
				SLA_school_x = SLAs_in_postcode_x[i];
				local_student_population = sla_numStudents_map[SLAs_in_postcode_x[i]];
				SLA_assignment_flag = true;
				std::cout << "SLA " << SLA_school_x << " assigned based on max-student rule" << std::endl;
			}
		}


		// now find a DZ in the selected SLA with enough workers
		DZs_in_SLA_x = sla_dz_map[SLA_school_x];

		bool DZ_assignment_flag = false;

		if (SLA_assignment_flag == true)
		{
			for (int i = 0; i < DZs_in_SLA_x.size(); i++)
			{
				num_teachers_required =
				ceil(school_schoolsize_map[schools_of_size_x.front()]/NUM_STUDENTS_TO_TEACHERS);

				if (dz_numWorkers_map[DZs_in_SLA_x[i]] > num_teachers_required)
				{
					DZ_school_x = DZs_in_SLA_x[i];
					dz_numWorkers_map[DZ_school_x] -= num_teachers_required;

					std::cout << "school " << schools_of_size_x.front() << " assigned to DZ "
					<< DZ_school_x << " which now has " << dz_numWorkers_map[DZ_school_x]
					<< " unassigned workers" << std::endl;

					DZ_assignment_flag = true;

					break;
				}
			}
			// if no DZs were found, randomly assign the SLA
			if (DZ_assignment_flag == false)
			{
				SLA_assignment_flag = false;
				std::cout << " could not find a DZ in the SLA found using max-student rule, reverting to stochastic approach" << std::endl;
			}
		}

		if (SLA_assignment_flag == false)
		{
			std::cout << "max-student approach failed, finding random sla" << std::endl;
			while (SLAs_in_postcode_x.size() > 0)
			{
				unsigned long selected_sla_index;
				unsigned long selected_SLA;

				// randomly find an SLA, and check for possible DZ assignments
				selected_sla_index = rand() % SLAs_in_postcode_x.size();//locates a random index
				selected_SLA = SLAs_in_postcode_x[selected_sla_index];
				DZs_in_SLA_x = sla_dz_map[selected_SLA];

				std::cout << "searching " << DZs_in_SLA_x.size() << " DZs " << "in SLA " << selected_SLA << std::endl;

				// look through DZs to see if there is one with enough workers
				for (int i = 0; i < DZs_in_SLA_x.size(); i++)
				{
					num_teachers_required =
					ceil(school_schoolsize_map[schools_of_size_x.front()]/NUM_STUDENTS_TO_TEACHERS);

					std::cout << "checking DZs for n workers. DZ:" << DZs_in_SLA_x[i]
					<< " contains " << dz_numWorkers_map[DZs_in_SLA_x[i]] << " workers." << std::endl;

					if (dz_numWorkers_map[DZs_in_SLA_x[i]] > num_teachers_required)
					{
						DZ_school_x = DZs_in_SLA_x[i];
						dz_numWorkers_map[DZ_school_x] -= num_teachers_required;

						std::cout << "school " << schools_of_size_x.front() << " assigned to DZ "
						<< DZ_school_x << " which now has " << dz_numWorkers_map[DZ_school_x]
						<< " unassigned workers" << std::endl;

						DZ_assignment_flag = true;

						break;
					}
				}

				if (DZ_assignment_flag == true)
				{
					SLA_school_x = selected_SLA;
					SLA_assignment_flag = true;
					std::cout << "SLA selected randomly: " << SLA_school_x << std::endl;
					break;
				}
				else
				{
					SLAs_in_postcode_x.erase(SLAs_in_postcode_x.begin() + selected_sla_index);
				}
			}
		}


		if (DZ_assignment_flag == false)
		{
			std::cout << "could not find a DZ for school " << schools_of_size_x.front() <<
			" in postcode " << postcode_school_x << std::endl;

			ACARA_Discrepency << postcode_school_x << " " << schools_of_size_x.front() << std::endl;

			std::cout << "Press enter to continue ...";
			std::cin.get();
		}


		// TODO: report error if there are not enough students in the postcode, sla, whatever, to populate the school

		// subtract students from local population
		if (sla_numStudents_map[SLA_school_x] >= school_schoolsize_map[schools_of_size_x.front()])
		{
			sla_numStudents_map[SLA_school_x] -= school_schoolsize_map[schools_of_size_x.front()];
		}
		else
		{
			sla_numStudents_map[SLA_school_x] = 0;
			std::cout << "WARNING: no more students in SLA " << SLA_school_x << std::endl;
		}

		if (SLA_assignment_flag == false)
		{
			std::cout << "could not find an SLA for school " << schools_of_size_x.front() <<
			" in postcode " << postcode_school_x << std::endl;
		}

		else
		{
			std::cout << "school " << schools_of_size_x.front() << " assigned to SLA "
			<< SLA_school_x << " which now has " << sla_numStudents_map[SLA_school_x]
			<< " unassigned students" << std::endl;


			// generate the school and add it to the output maps
			//sla_type sla = dz_sla_map.at( selectedDZ );
			// Schools with no DZ or SLA will not be constructed
			struct School * cschool = new struct School;
			cschool->ID = schools_of_size_x.front();
			cschool->max_students = size_school_map_it->first; // will have to change this to a bin size
			cschool->num_students = 0;
			cschool->SLA = SLA_school_x;
			cschool->DZ = DZ_school_x;

			school_id_map[cschool->ID] = cschool;
			sla_school_map[SLA_school_x].push_back( cschool );


		}


		std::cout<<"\n"<<std::endl;

		// remove the school from the list
		schools_of_size_x.erase(schools_of_size_x.begin());
	}
}

ACARA_Discrepency.close();

}
