/*
   *Copyright:  

   ACEMod (2018--2021), The University of Sydney's invention disclosure CDIP Ref. 2019-123.
   Copyright ©  Oliver Cliff, Cameron Zachreson, Nathan Harding, Kristopher Fair, Mikhail Prokopenko, Mahendarajah Piraveenan

   AMTraC-19 (2020-2021), The University of Sydney's invention disclosure CDIP 2020-018 (2020).
   Copyright ©  Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   AMTraC-19 (2022-2023), The University of Sydney.
   Copyright ©  Sheryl Chang, Quang Dang Nguyen, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   AMTraC-19 (2024), The University of Sydney.
   Copyright ©  Quang Dang Nguyen, Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   *EndCopyright:
*/

#ifndef UTILS_HPP_
#define UTILS_HPP_

#include <map>
#include <algorithm>
#include <cmath>
#include <boost/math/special_functions/erf.hpp>

#include <string>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <iostream>

/* Return cumulative probability of a half-normal distribution at x (st dev sigma) */
inline float foldedNormCDF( float sigma, float x ) {
	return boost::math::erf( x / sqrt( 2 ) / sigma );
}

/* Return the value of the inverse CDF of a half-normal distribution at probability p (st dev sigma) */
inline float invFoldedNormCDF( float sigma, float p ) {
    p = ( p < 1e-4 ) ? 1e-4 : p;
    p = ( p > 0.999 ) ? 0.9999 : p;
    return sigma * sqrt( 2 ) * boost::math::erf_inv( p );
}

/* Generate CDF from some unnormalised prob distribution (dist) and the bins */
template<class T> std::vector< std::pair< T, float > > inline generateCDF( const std::vector< T > bins, const std::vector< float > dist ) {  
    std::vector< std::pair< T, float > > cdf;
    /* Cumulate distribution... */
    float sum = 0; 
    for ( unsigned i = 0; i < dist.size(); ++i ) {
        sum += dist.at(i);
        cdf.emplace_back( bins.at(i), sum );
    }
   
    /* ...and normalise it */
    for ( auto& outcome : cdf ) {
        outcome.second /= sum;
    }
  
    return cdf;
}

template<class T> T sampleCDF( const std::vector< std::pair< T, float > >& cdf ) {
    float rval = rand() / float( RAND_MAX );
    for ( const auto& outcome : cdf ) {
        if ( rval <= outcome.second ) {
            return outcome.first;
        }
    }
    return cdf.rbegin()->first;
}

#endif //UTILS_HPP_