/*
   *Copyright:  

   ACEMod (2018--2021), The University of Sydney's invention disclosure CDIP Ref. 2019-123.
   Copyright ©  Oliver Cliff, Cameron Zachreson, Nathan Harding, Kristopher Fair, Mikhail Prokopenko, Mahendarajah Piraveenan

   AMTraC-19 (2020-2021), The University of Sydney's invention disclosure CDIP 2020-018 (2020).
   Copyright ©  Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   AMTraC-19 (2022-2023), The University of Sydney.
   Copyright ©  Sheryl Chang, Quang Dang Nguyen, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   AMTraC-19 (2024), The University of Sydney.
   Copyright ©  Quang Dang Nguyen, Sheryl Chang, Nathan Harding, Cameron Zachreson, Oliver Cliff, Mikhail Prokopenko

   *EndCopyright:
*/

#ifndef  VIRUS_H_
#define  VIRUS_H_

class Virus {
    public:
        Virus( 
            float trans_scaler_mean, float trans_scaler_std,
            float latent_period_mean, float latent_period_std, 
            float incubation_period_mu, float incubation_period_sig, 
            float recovery_period_hi, float recovery_period_low, 
            float after_recovery_period_mean, float after_recovery_period_std,
            float adult_symptomatic_fraction, float child_symptomatic_fraction, 
            float asymptomatic_infectivity,
            unsigned origin_code
        );
    
        Virus(void);

        void set_trans_scaler_dist(float, float);
        void set_latent_period_dist(float, float);
        void set_incubation_period_dist(float, float);
        void set_recovery_period_dist(float, float);
        void set_after_recovery_period_dist(float, float);
        void set_adult_symptomatic_fraction(float);
        void set_child_symptomatic_fraction(float);
        void set_asymptomatic_infectivity(float);
        
        void set_origin_code(unsigned);

        float sample_trans_scaler(void);
        float sample_latent_period(void);
        float sample_incubation_period(void);
        int   sample_recovery_period(void);
        float sample_after_recovery_period(void);

        float get_trans_scaler_mean(void);
        float get_trans_scaler_std(void);
        float get_latent_period_mean(void);
        float get_latent_period_std(void);
        float get_after_recovery_period_mean(void);
        float get_after_recovery_period_std(void);
        float get_adult_symptomatic_fraction(void);
        float get_child_symptomatic_fraction(void);
        float get_asymptomatic_infectivity(void);
        float get_incubation_period_dist_mu(void);
        float get_incubation_period_dist_sig(void);
        float get_recovery_period_dist_high(void);
        float get_recovery_period_dist_low(void);

        unsigned get_origin_code(void);

        void output_viral_info(void);

    private:

        float trans_scaler_mean, trans_scaler_std;
        float latent_period_mean, latent_period_std;
        float incubation_period_mu, incubation_period_sig, incubation_period;
        float recovery_period_hi, recovery_period_low, recovery_period;
        float after_recovery_period_mean, after_recovery_period_std;
        float adult_symptomatic_fraction, child_symptomatic_fraction, asymptomatic_infectivity;

        unsigned origin_code;    // name convention for the original strain. e.g.: 0: Ancestral, 1: Delta, 2: Omicron
                                 // for book keeping purpose only.
};

#endif